#ifndef EVENT_LOOP_H
#define EVENT_LOOP_H

/**
 * This class implements an event loop.  This class is probably only
 * functional on unix-like operating systems.  This is why it is not
 * part of cgul. <p>
 *
 * NOTE: This class is not thread-safe.  This is by design because you
 * generally only call this class's methods from the same thread.  If
 * you need to call this class's methods from multiple threads, you'll
 * need to add external locking. <p>
 *
 * NOTE: This class is re-entrant.  This allows any callback to
 * directly add or remove any callback including itself.
 *
 * @file
 * @author Paul Serice
 * @brief event loop implementation
 */

#include <signal.h>
#include "cgul_exception.h"
#include "event_loop_common.h"

CGUL_BEGIN_C

/*************************************************************************
 * Opaque Pointers
 *************************************************************************/

/** Opaque pointer to a <code>event_loop</code> instance. */
typedef struct event_loop* event_loop_t;

/** Opaque pointer to a timeout handle. */
typedef void* event_loop__timeout_t;

/*************************************************************************
 * C Callback Interface
 *************************************************************************/

/**
 * File descriptor ready callback.  Functions you write that match
 * this prototype can be registered to be called when the file
 * descriptor <code>fd</code> is ready.  The <code>data</code>
 * parameter is the usual client data.  The <code>cex</code> parameter
 * is a pointer that can be used to throw a c-style exception.  Any
 * exception that escapes this callback will be caught by the main
 * event loop and printed to standard error.  If you do not like this
 * behavior, you do not have to use <code>cex</code>, or you can catch
 * it inside your function.
 *
 * @param[in,out] cex c-style exception
 * @param[in] fd file descriptor
 * @param[in] data client data
 */
typedef void (*event_loop__fd_ready_cb_t)(cgul_exception_t* cex,
                                          int fd,
                                          void* data);

/**
 * Timeout expired callback.  Functions you write that match this
 * prototype can be registered to be called when a timeout expires.
 * The <code>timeout</code> parameter is the same value returned by
 * <code>event_loop__add_timeout()</code>.  It can be used to remove
 * the timeout so that its callback will no longer be called.  The
 * <code>data</code> parameter is the usual client data.  The
 * <code>cex</code> parameter is a pointer that can be used to throw a
 * c-style exception.  Any exception that escapes this callback will
 * be caught by the main event loop and printed to standard error.  If
 * you do not like this behavior, you do not have to use
 * <code>cex</code>, or you can catch it inside your function.
 *
 * @param[in,out] cex c-style exception
 * @param[in] timeout timeout handle
 * @param[in] data client data
 */
typedef void (*event_loop__timeout_cb_t)(cgul_exception_t* cex,
                                         event_loop__timeout_t timeout,
                                         void* data);

/**
 * Posix asynchronous signal caught callback.  Functions you write
 * that match this prototype can be registered to be called when a
 * particular signal has been caught.  The <code>signal_number</code>
 * is the signal number for the signal that was caught.  The
 * <code>data</code> parameter is the usual client data.  The
 * <code>cex</code> parameter is a pointer that can be used to throw a
 * c-style exception.  Any exception that escapes this callback will
 * be caught by the main event loop and printed to standard error.  If
 * you do not like this behavior, you do not have to use
 * <code>cex</code>, or you can catch it inside your function.
 *
 * @param[in,out] cex c-style exception
 * @param[in] signal_number signal number for the signal that was caught
 * @param[in] data client data
 */
typedef void (*event_loop__signal_cb_t)(cgul_exception_t* cex,
                                        int signal_number,
                                        void* data);

/*************************************************************************
 * C++ Listener Interface
 *************************************************************************/

#ifdef __cplusplus

/**
 * To have <code>event_loop</code> directly invoke C++ member methods,
 * derive your class from the <code>event_loop_listener</code> class.
 * All you need to do then is to declare any member method that
 * implements the correct function signature for an event listener.
 * Then, use one of the following macros to upcast a "pointer to
 * member function" to an instance of the class and pass the result
 * into one of the <code>event_loop__add_*_listener()</code> methods
 * below: <p>
 *
 * <code>
 *     EVENT_LOOP__UPCAST_FD_READY()
 *     EVENT_LOOP__UPCAST_TIMEOUT()
 *     EVENT_LOOP__UPCAST_SIGNAL_HANDLER()
 * </code>
 *
 * This works because it is possible to upcast a "pointer to member
 * function" using <code>static_cast</code> [sic] (hence the macros).
 * This lets the <code>event_router</code> class invoke member methods
 * of derived classes.  It can even invoke non-virtual methods of
 * derived classes (which is why <code>event_loop_listener</code> does
 * not need to define any virtual methods allowing you to name the
 * methods whatever you like and allowing you to create as many as you
 * need). <p>
 *
 * The one gotcha is that the C++ standard places restrictions on when
 * it is possible to upcast a "pointer to member function."  If you
 * stick with single inheritance or use
 * <code>event_loop_listener</code> as a "mixin", you should be ok.
 * If you are having trouble getting the provided macros to compile,
 * you have probably hit one of these exceptions.  In that case, just
 * use a <code>static</code> C++ method (which can be passed into
 * <code>event_loop</code> as a C-style function) as an adapter that
 * can invoke the member method. <p>
 *
 * The rationale for this is that <code>std::mem_fun</code> is
 * deprecated and <code>std::function</code> which was introduced in
 * C++11 is not widely available yet.  Furthermore, write our own
 * version of <code>std::mem_fun</code> is not going to work because
 * it would require using C++ templates which requires all the code to
 * be exposed in the header which violates C-style encapsulation. <p>
 *
 * Lastly, we could have this class declare virtual methods that
 * implement the necessary callbacks, but this, in general, would
 * require creating a new object for each event listener which
 * requires is more boilerplate than just using C-style functions.
 * The advantage of how it is actually done is that any member
 * function that implements the correct interface can be used as an
 * event listener.  Thus, if you need 20 event listeners, you can have
 * one class with 20 methods instead of 20 class with one method each.
 * Both would work, but the former should be much easier to
 * implement. <p>
 */
class event_loop_listener {
public:

    /**
     * Default Constructor
     */
    event_loop_listener()
    {
        /* empty */
    }

    /**
     * Destructor
     */
    virtual
    ~event_loop_listener()
    {
        /* empty */
    }
};

/**
 * Upcast a "pointer to member function" <code>ptr_to_mem</code> for a
 * member method that is part of a class derived from
 * <code>event_loop_listener</code> so it can be passed into any of
 * the following functions for adding C++ listeners: <p>
 *
 * <pre>
 *   event_loop__add_read_fd_listener()
 *   event_loop__add_write_fd_listener()
 *   event_loop__add_exception_fd_listener()
 * </pre>
 *
 * For example, assume your derived class is declared as follows: <p>
 *
 * <code>
 *   class my_class : public event_loop_listener {
 *   public:
 *       void my_read_fd(int fd);
 *   };
 * </code>
 *
 * You can then upcast <code>my_read_fd</code> as follows: <p>
 *
 * <code>
 *   EVENT_LOOP__UPCAST_FD_READY(&my_class::my_read_fd)
 * </code>
 *
 * @param[in] ptr_to_mem pointer to member method
 * @see event_loop_listener
 */
#define EVENT_LOOP__UPCAST_FD_READY(ptr_to_mem) \
  static_cast<void (event_loop_listener::*)(int)>(ptr_to_mem)

/**
 * Upcast a "pointer to member function" <code>ptr_to_mem</code> for a
 * member method that is part of a class derived from
 * <code>event_loop_listener</code> so it can be passed into the
 * <code>event_loop__add_timeout_listener()</code> function for adding
 * C++ listeners. <p>
 *
 * For example, assume your derived class is declared as follows: <p>
 *
 * <code>
 *   class my_class : public event_loop_listener {
 *   public:
 *       void my_timeout(event_loop__timeout_t timeout);
 *   };
 * </code>
 *
 * You can then upcast <code>my_timeout</code> as follows: <p>
 *
 * <code>
 *   EVENT_LOOP__UPCAST_TIMEOUT(&my_class::my_timeout)
 * </code>
 *
 * @param[in] ptr_to_mem pointer to member method
 * @see event_loop_listener
 */
#define EVENT_LOOP__UPCAST_TIMEOUT(ptr_to_mem) \
  static_cast<void (event_loop_listener::*)(event_loop__timeout_t)>(ptr_to_mem)

/**
 * Upcast a "pointer to member function" <code>ptr_to_mem</code> for a
 * member method that is part of a class derived from
 * <code>event_loop_listener</code> so it can be passed into the
 * <code>event_loop__add_signal_handler_listener()</code> function for
 * adding C++ listeners. <p>
 *
 * For example, assume your derived class is declared as follows: <p>
 *
 * <code>
 *   class my_class : public event_loop_listener {
 *   public:
 *       void my_signal_handler(int signal_number);
 *   };
 * </code>
 *
 * You can then upcast <code>my_signal_handler</code> as follows: <p>
 *
 * <code>
 *   EVENT_LOOP__UPCAST_SIGNAL_HANDLER(&my_class::my_signal_handler)
 * </code>
 *
 * @param[in] ptr_to_mem pointer to member method
 * @see event_loop_listener
 */
#define EVENT_LOOP__UPCAST_SIGNAL_HANDLER(ptr_to_mem) \
  static_cast<void (event_loop_listener::*)(int)>(ptr_to_mem)

#endif  /*  __cplusplus  */

/*************************************************************************
 * event_loop
 *************************************************************************/

/**
 * Get the single instance of the <code>event_loop</code> class.  The
 * single instance is created the first time this method is called.
 * If the instance cannot be created, an exception is thrown. <p>
 *
 * @note For the main thread, use the singleton returned by this
 * method which can handle Posix asynchronous signals.  Do not use a
 * per-thread instance returned by <code>event_loop__new()</code>
 * which cannot handle Posix asynchronous signals. <p>
 *
 * @param[in,out] cex c-style exception
 * @return single instance (i.e., main event loop)
 */
extern event_loop_t
event_loop__get_instance(cgul_exception_t* cex);

/**
 * Put the single instance of the <code>event_loop</code> class.  The
 * single instance is destroyed whenever this method is called.  In
 * general, it should only be called before exiting the application if
 * it is called at all.
 */
extern void
event_loop__put_instance();

/**
 * Create a new, per-thread <code>event_loop</code> object that
 * handles file descriptor and timeout events but not Posix
 * asynchronous signal events.  The caller is responsible for freeing
 * the object by calling <code>event_loop_delete()</code>.  If memory
 * cannot be allocated, <code>NULL</code> is returned, and an
 * exception is thrown. <p>
 *
 * @note For the main thread, do not create a new
 * <code>event_loop</code> object using this method; instead, use the
 * singleton returned by <code>event_loop__get_instance()</code> which
 * has global visibility and can handle Posix asynchronous
 * signals. <p>
 *
 * @param[in,out] cex c-style exception
 * @return new <code>event_loop</code> instance
 */
extern event_loop_t
event_loop__new(cgul_exception_t* cex);

/**
 * This method frees all internally allocated memory associated with
 * <code>loop</code>.  Do not try to access <code>loop</code> after
 * calling this method. <p>
 *
 * @note Only call this method on instances returned by direct calls
 * to <code>event_loop__new()</code>.  Do not call this method on the
 * singleton returned by <code>event_loop__get_instance()</code>;
 * instead call <code>event_loop__put_instance()</code> to properly
 * delete the singleton. <p>
 *
 * @param[in] loop <code>event_loop</code> instance
 */
extern void
event_loop__delete(event_loop_t loop);

/**
 * This method starts the main event loop which only returns when one
 * of the callbacks registered with the event loop invokes
 * <code>event_loop__stop()</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 */
extern void
event_loop__start(cgul_exception_t* cex,
                  event_loop_t loop);

/**
 * Call this method from within an event callback or from a different
 * thread in order to stop processing events.  This will cause
 * <code>event_loop__start</code> to return to its caller.  This
 * method is thread-safe.  If an error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 */
extern void
event_loop__stop(cgul_exception_t* cex,
                 event_loop_t loop);

/**
 * This method arranges for <code>cb</code> to be called synchronously
 * with <code>cb_data</code> when a Posix asynchronous signal is
 * caught.  The big advantage of this is that <code>cb</code> runs in
 * the synchronous context of the event loop's thread instead of in
 * the asynchronous context of a true signal handler.  If
 * <code>cb</code> is <code>NULL</code>, the signal is effectively
 * ignored.  If the signal cannot be registered, an exception is
 * thrown. <p>
 *
 * You can change <code>cb</code> and <code>cb_data</code> for a
 * particular signal by calling this method more than once, but only
 * one signal handler can simultaneously be registered per signal. <p>
 *
 * As a convenience, the <code>event_loop</code> class is re-entrant
 * with respect to the event loop's single thread of execution.  This
 * means you can call any of the methods of this class even from
 * callbacks registered with this class. <p>
 *
 * If an error occurs, <code>NULL</code> is returned, and an exception
 * is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] signal_number signal number
 * @param[in] cb callback
 * @param[in] cb_data callback data passed to <code>cb</code>
 */
extern void
event_loop__add_signal_handler(cgul_exception_t* cex,
                               event_loop_t loop,
                               int signal_number,
                               event_loop__signal_cb_t cb,
                               void* cb_data);

#ifdef __cplusplus

/**
 * This method arranges for the C++ event listener
 * <code>listener</code> to be called synchronously on the object
 * <code>obj</code> when a Posix asynchronous signal is caught.  The
 * big advantage of this is that <code>listener</code> runs in the
 * synchronous context of the event loop's thread instead of in the
 * asynchronous context of a true signal handler.  If the signal
 * cannot be registered, an exception is thrown. <p>
 *
 * You can change <code>listener</code> and <code>obj</code> for a
 * particular signal by calling this method more than once, but only
 * one signal handler can simultaneously be registered per signal. <p>
 *
 * When the Posix asynchronous signal is synchronously detected, the
 * C++ "pointer to member function" <code>listener</code> is invoked
 * on the object <code>obj</code> which must be an instance of a class
 * derived from <code>event_loop_listener</code>.  Use the
 * <code>EVENT_LOOP__UPCAST_SIGNAL_HANDLER()</code> macro to properly
 * upcast your "pointer to member function" for your derived class to
 * a type of "pointer to member function" expected by this function.
 * This should look something like the following: <p>
 *
 *   <code>
 *       event_loop__add_signal_handler_listener(
 *           cex, loop, signal_number, obj,
 *           EVENT_LOOP__UPCAST_SIGNAL_HANDLER(&my_class::my_signal_handler));
 *   </code>
 *
 * As a convenience, the <code>event_loop</code> class is re-entrant
 * with respect to the event loop's single thread of execution.  This
 * means you can call any of the methods of this class even from
 * callbacks registered with this class. <p>
 *
 * If an error occurs, <code>NULL</code> is returned, and an exception
 * is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] signal_number signal number
 * @param[in] obj object that implements the event listener
 * @param[in] listener event listener
 * @see event_loop_listener
 * @see EVENT_LOOP__UPCAST_SIGNAL_HANDLER
 */
extern void
event_loop__add_signal_handler_listener(
    cgul_exception_t* cex,
    event_loop_t loop,
    int signal_number,
    event_loop_listener* obj,
    void (event_loop_listener::*listener)(int));

#endif

/**
 * Remove the the signal handler for the signal
 * <code>signal_number</code> from the list of Posix asynchronous
 * signals handled by the event loop.  This causes the default system
 * action for the signal to be restored.  If an error occurs, an
 * exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] signal_number signal number
 */
extern void
event_loop__remove_signal_handler(cgul_exception_t* cex,
                                  event_loop_t loop,
                                  int signal_number);

/**
 * This method adds a timeout event that repeatedly schedules itself
 * every <code>period</code> seconds.  Because <code>period</code> is
 * a double, timeouts with millisecond resolution can easily be
 * scheduled.  The only guarantee regarding when the timeout occurs is
 * that it occurs at least <code>period</code> seconds after it has
 * been scheduled.  If <code>period</code> is less than zero, the
 * timeout will not be registered. <p>
 *
 * The <code>event_loop</code> class is re-entrant with respect to the
 * event loop's single thread of execution making it possible to
 * remove a timeout from <code>cb</code> itself. <p>
 *
 * The value returned is the handle for the timeout.  It uniquely
 * identifies the timeout and must be passed into
 * <code>event_loop__remove_timeout()</code> in order to remove the
 * timeout.  If you plan on removing the callback from inside
 * <code>cb</code>, you do not need to keep a reference to the value
 * that is returned as it is always passed into your callback. <p>
 *
 * If an error occurs, <code>NULL</code> is returned, and an exception
 * is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] period timeout period
 * @param[in] cb_data callback data passed to <code>cb</code>
 * @param[in] cb callback
 * @return timeout handle
 */
extern event_loop__timeout_t
event_loop__add_timeout(cgul_exception_t* cex,
                        event_loop_t loop,
                        double period,
                        event_loop__timeout_cb_t cb,
                        void* cb_data);

#ifdef __cplusplus

/**
 * This method adds a C++ timeout event listener that repeatedly
 * schedules itself every <code>period</code> seconds.  Because
 * <code>period</code> is a double, timeouts with millisecond
 * resolution can easily be scheduled.  The only guarantee regarding
 * when the timeout occurs is that it occurs at least
 * <code>period</code> seconds after it has been scheduled.  If
 * <code>period</code> is less than zero, the timeout will not be
 * registered. <p>
 *
 * When the timeout expires, the C++ "pointer to member function"
 * <code>listener</code> is invoked on the object <code>obj</code>
 * which must be an instance of a class derived from
 * <code>event_loop_listener</code>.  Use the
 * <code>EVENT_LOOP__UPCAST_TIMEOUT()</code> macro to properly upcast
 * your "pointer to member function" for your derived class to a type
 * of "pointer to member function" expected by this function.  This
 * should look something like the following: <p>
 *
 *   <code>
 *       event_loop__add_timeout_listener(
 *           cex, loop, period, obj,
 *           EVENT_LOOP__UPCAST_TIMEOUT(&my_class::my_timeout));
 *   </code>
 *
 * The <code>event_loop</code> class is re-entrant with respect to the
 * event loop's single thread of execution making it possible to
 * remove a timeout from <code>listener</code> itself. <p>
 *
 * The value returned is the handle for the timeout.  It uniquely
 * identifies the timeout and must be passed into
 * <code>event_loop__remove_timeout()</code> in order to remove the
 * timeout.  If you plan on removing the listener from inside
 * <code>listener</code>, you do not need to keep a reference to the value
 * that is returned as it is always passed into your listener. <p>
 *
 * If an error occurs, <code>NULL</code> is returned, and an exception
 * is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] period timeout period
 * @param[in] obj object that implements the event listener
 * @param[in] listener event listener
 * @return timeout handle
 * @see event_loop_listener
 * @see EVENT_LOOP__UPCAST_TIMEOUT
 */
extern event_loop__timeout_t
event_loop__add_timeout_listener(
    cgul_exception_t* cex,
    event_loop_t loop,
    double period,
    event_loop_listener* obj,
    void (event_loop_listener::*listener)(event_loop__timeout_t));

#endif

/**
 * This function removes timeouts registered with
 * <code>event_loop__add_timeout()</code>.  After calling this
 * function, <code>timeout</code> is invalid.  If an error occurs, an
 * exception is thrown. <p>
 *
 * This method can throw an out-of-memory exception.  This can happen
 * when this method is called from any timeout callback because this
 * causes this class to be re-entered which requires that the request
 * to remove the timeout be added to a queue so it can be honored
 * later when it is safe to do so. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] timeout timeout handle
 */
extern void
event_loop__remove_timeout(cgul_exception_t* cex,
                           event_loop_t loop,
                           event_loop__timeout_t timeout);

/**
 * This method adds <code>fd</code> to the list of file descriptors
 * the event loop monitors.  The event loop will then call
 * <code>cb</code> with the client-supplied <code>cb_data</code>
 * pointer when <code>fd</code> is ready to be read. <p>
 *
 * You can change <code>cb</code> and <code>cb_data</code> for a
 * particular file descriptor by calling this method more than once,
 * but only one "read" callback per file descriptor can simultaneously
 * be registered. <p>
 *
 * As a convenience, the <code>event_loop</code> class is re-entrant
 * with respect to the event loop's single thread of execution.  This
 * means you can call any of the methods of this class even from
 * callbacks registered with this class. <p>
 *
 * If an error occurs, an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] fd file descriptor to monitor
 * @param[in] cb callback
 * @param[in] cb_data callback data passed to <code>cb</code>
 */
extern void
event_loop__add_read_fd(cgul_exception_t* cex,
                        event_loop_t loop,
                        int fd,
                        event_loop__fd_ready_cb_t cb,
                        void* cb_data);

#ifdef __cplusplus

/**
 * This method adds <code>fd</code> to the list of file descriptors
 * the event loop monitors.  You can change <code>listener</code> and
 * <code>obj</code> for a particular file descriptor by calling this
 * method more than once, but only one "read" event listener per file
 * descriptor can simultaneously be registered. <p>
 *
 * When the file descriptor is ready, the C++ "pointer to member
 * function" <code>listener</code> is invoked on the object
 * <code>obj</code> which must be an instance of a class derived from
 * <code>event_loop_listener</code>.  Use the
 * <code>EVENT_LOOP__UPCAST_FD_READY()</code> macro to properly upcast
 * your "pointer to member function" for your derived class to a type
 * of "pointer to member function" expected by this function.  This
 * should look something like the following: <p>
 *
 *   <code>
 *       event_loop__add_read_fd_listener(
 *           cex, loop, fd, obj,
 *           EVENT_LOOP__UPCAST_FD_READY(&my_class::my_listener));
 *   </code>
 *
 * As a convenience, the <code>event_loop</code> class is re-entrant
 * with respect to the event loop's single thread of execution.  This
 * means you can call any of the methods of this class even from event
 * listeners registered with this class. <p>
 *
 * If an error occurs, an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] fd file descriptor to monitor
 * @param[in] obj object that implements the event listener
 * @param[in] listener event listener
 * @see event_loop_listener
 * @see EVENT_LOOP__UPCAST_FD_READY
 */
extern void
event_loop__add_read_fd_listener(
    cgul_exception_t* cex,
    event_loop_t loop,
    int fd,
    event_loop_listener* obj,
    void (event_loop_listener::*listener)(int));

#endif  /*  __cplusplus  */

/**
 * Remove <code>fd</code> from the list of file descriptors the event
 * loop monitors to see when they are ready to be read.  If an error
 * occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] fd file descriptor to remove
 */
extern void
event_loop__remove_read_fd(cgul_exception_t* cex,
                           event_loop_t loop,
                           int fd);

/**
 * This method adds <code>fd</code> to the list of file descriptors
 * the event loop monitors.  The event loop will then call
 * <code>cb</code> with the client-supplied <code>cb_data</code>
 * pointer when <code>fd</code> is ready to be written. <p>
 *
 * You can change <code>cb</code> and <code>cb_data</code> for a
 * particular file descriptor by calling this method more than once,
 * but only one "write" callback per file descriptor can
 * simultaneously be registered. <p>
 *
 * As a convenience, the <code>event_loop</code> class is re-entrant
 * with respect to the event loop's single thread of execution.  This
 * means you can call any of the methods of this class even from
 * callbacks registered with this class. <p>
 *
 * If an error occurs, an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] fd file descriptor to monitor
 * @param[in] cb callback
 * @param[in] cb_data callback data passed to <code>cb</code>
 */
extern void
event_loop__add_write_fd(cgul_exception_t* cex,
                         event_loop_t loop,
                         int fd,
                         event_loop__fd_ready_cb_t cb,
                         void* cb_data);

#ifdef __cplusplus

/**
 * This method adds <code>fd</code> to the list of file descriptors
 * the event loop monitors.  You can change <code>listener</code> and
 * <code>obj</code> for a particular file descriptor by calling this
 * method more than once, but only one "write" event listener per file
 * descriptor can simultaneously be registered. <p>
 *
 * When the file descriptor is ready, the C++ "pointer to member
 * function" <code>listener</code> is invoked on the object
 * <code>obj</code> which must be an instance of a class derived from
 * <code>event_loop_listener</code>.  Use the
 * <code>EVENT_LOOP__UPCAST_FD_READY()</code> macro to properly upcast
 * your "pointer to member function" for your derived class to a type
 * of "pointer to member function" expected by this function.  This
 * should look something like the following: <p>
 *
 *   <code>
 *       event_loop__add_write_fd_listener(
 *           cex, loop, fd, obj,
 *           EVENT_LOOP__UPCAST_FD_READY(&my_class::my_listener));
 *   </code>
 *
 * As a convenience, the <code>event_loop</code> class is re-entrant
 * with respect to the event loop's single thread of execution.  This
 * means you can call any of the methods of this class even from event
 * listeners registered with this class. <p>
 *
 * If an error occurs, an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] fd file descriptor to monitor
 * @param[in] obj object that implements the event listener
 * @param[in] listener event listener
 * @see event_loop_listener
 * @see EVENT_LOOP__UPCAST_FD_READY
 */
extern void
event_loop__add_write_fd_listener(
    cgul_exception_t* cex,
    event_loop_t loop,
    int fd,
    event_loop_listener* obj,
    void (event_loop_listener::*listener)(int));

#endif  /*  __cplusplus  */

/**
 * Remove <code>fd</code> from the list of file descriptors the event
 * loop monitors to see when they are ready to be written.  If an
 * error occurs, an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] fd file descriptor to remove
 */
extern void
event_loop__remove_write_fd(cgul_exception_t* cex,
                            event_loop_t loop,
                            int fd);

/**
 * This method adds <code>fd</code> to the list of file descriptors
 * the event loop monitors.  The event loop will then call
 * <code>cb</code> with the client-supplied <code>cb_data</code>
 * pointer when <code>fd</code> has an exception. <p>
 *
 * You can change <code>cb</code> and <code>cb_data</code> for a
 * particular file descriptor by calling this method more than once,
 * but only one "exception" callback per file descriptor can
 * simultaneously be registered. <p>
 *
 * As a convenience, the <code>event_loop</code> class is re-entrant
 * with respect to the event loop's single thread of execution.  This
 * means you can call any of the methods of this class even from
 * callbacks registered with this class. <p>
 *
 * If an error occurs, an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] fd file descriptor to monitor
 * @param[in] cb callback
 * @param[in] cb_data callback data passed to <code>cb</code>
 */
extern void
event_loop__add_exception_fd(cgul_exception_t* cex,
                             event_loop_t loop,
                             int fd,
                             event_loop__fd_ready_cb_t cb,
                             void* cb_data);

#ifdef __cplusplus

/**
 * This method adds <code>fd</code> to the list of file descriptors
 * the event loop monitors.  You can change <code>listener</code> and
 * <code>obj</code> for a particular file descriptor by calling this
 * method more than once, but only one "exception" event listener per
 * file descriptor can simultaneously be registered. <p>
 *
 * When the file descriptor is ready, the C++ "pointer to member
 * function" <code>listener</code> is invoked on the object
 * <code>obj</code> which must be an instance of a class derived from
 * <code>event_loop_listener</code>.  Use the
 * <code>EVENT_LOOP__UPCAST_FD_READY()</code> macro to properly upcast
 * your "pointer to member function" for your derived class to a type
 * of "pointer to member function" expected by this function.  This
 * should look something like the following: <p>
 *
 *   <code>
 *       event_loop__add_exception_fd_listener(
 *           cex, loop, fd, obj,
 *           EVENT_LOOP__UPCAST_FD_READY(&my_class::my_listener));
 *   </code>
 *
 * As a convenience, the <code>event_loop</code> class is re-entrant
 * with respect to the event loop's single thread of execution.  This
 * means you can call any of the methods of this class even from event
 * listeners registered with this class. <p>
 *
 * If an error occurs, an exception is thrown. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] fd file descriptor to monitor
 * @param[in] obj object that implements the event listener
 * @param[in] listener event listener
 * @see event_loop_listener
 * @see EVENT_LOOP__UPCAST_FD_READY
 */
extern void
event_loop__add_exception_fd_listener(
    cgul_exception_t* cex,
    event_loop_t loop,
    int fd,
    event_loop_listener* obj,
    void (event_loop_listener::*listener)(int));

#endif  /*  __cplusplus  */

/**
 * Remove <code>fd</code> from the list of file descriptors the event
 * loop monitors to see when an exception occurs.  If an error occurs,
 * an exception is thrown.
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] fd file descriptor to remove
 */
extern void
event_loop__remove_exception_fd(cgul_exception_t* cex,
                                event_loop_t loop,
                                int fd);

CGUL_END_C

#endif  /*  EVENT_LOOP_H  */
