#include <fcntl.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "event_loop_p.h"
#ifdef __cplusplus
#  include <stdexcept>
using namespace std;
#endif  /*  __cplusplus  */

/**
 * The single instance of this class.
 */
static event_loop_t event_loop__instance = NULL;

/**
 * Posix signal handlers do not allow client data to be passed into
 * the asynchronous signal handlers.  Thus, the
 * event_loop__async_to_sync_pipe is "static" so that both
 * event_loop__get_instance() and
 * event_loop__async_to_sync_signal_handler() can see it.  This is the
 * only reason I decided to make the event_loop class a singleton.  I
 * was worried about the interaction among multiple event_loop
 * instances.
 *
 * However, you can create and destroy per-thread instances of this
 * class using <code>event_loop__new()</code> and
 * <code>event_loop__delete()</code>.  You just will not be able to
 * receive events related to Posix asynchronous signals that have been
 * caught on the per-thread event loops; only, the event loop
 * associated with the singleton can handle asynchronous signals.
 */
static int event_loop__async_to_sync_pipe[2] = {-1, -1};

/**
 * Internal "stop" message.  This is written to the internal pipe
 * which allows <code>event_loop__stop()</code> to be thread-safe
 * because it only needs to write to the internal pipe instead of
 * modifying class data.  When the internal pipe is ready, the
 * <code>event_loop__read_internal_message()</code> function will be
 * called on the correct thread, and it can safely modify the state of
 * the <code>event_loop</code> instance.
 */
#define EVENT_LOOP__INTERNAL_STOP_MESSAGE (1)

/**
 * This function is used as the comparison function by
 * <code>loop->timeout_tree</code> tree to keep its keys sorted by the
 * expiration and then by the address of the key itself.  We use the
 * address of the key itself as part of the sorting criteria to insure
 * that all of our keys are unique. <p>
 *
 * This function returns <code>-1</code>, <code>0</code>, or
 * <code>1</code> to indicate whether <code>key1</code> is less than,
 * equal to, or greater than <code>key2</code> respectively.
 *
 * @param[in] key1 first key
 * @param[in] key2 second key
 * @return comparison of <code>key1</code> with <code>key2</code>
 */
static int
event_loop__compare_timeouts(const void* key1,
                             const void* key2)
{
    /* Assume equality. */
    int rv = 0;
    event_loop__timeout_data_t data1 = NULL;
    event_loop__timeout_data_t data2 = NULL;

    assert(key1);
    assert(key2);

    data1 = (event_loop__timeout_data_t)key1;
    data2 = (event_loop__timeout_data_t)key2;

    /* First, compare by expiration ("timeout"). */
    if (data1->timeout < data2->timeout) {
        rv = -1;
    } else if (data1->timeout > data2->timeout) {
        rv = 1;

    /* Then, compare by the address of the keys. */
    } else if (data1 < data2) {
        rv = -1;
    } else if (data1 > data2) {
        rv = 1;
    }

    return rv;
}

/**
 * This function is used as the comparison function by
 * <code>loop->signal_tree</code> tree to keep its keys sorted by
 * signal number. <p>
 *
 * This function returns less than zero, zero, or greater than zero to
 * indicate whether <code>key1</code> is less than, equal to, or
 * greater than <code>key2</code> respectively.
 *
 * @param[in] key1 first key
 * @param[in] key2 second key
 * @return comparison of <code>key1</code> with <code>key2</code>
 */
static int
event_loop__compare_signals(const void* key1,
                            const void* key2)
{
    /* Assume equality. */
    int rv = 0;
    event_loop__signal_data_t data1 = NULL;
    event_loop__signal_data_t data2 = NULL;

    assert(key1);
    assert(key2);

    data1 = (event_loop__signal_data_t)key1;
    data2 = (event_loop__signal_data_t)key2;

    /* Compare the signal numbers. */
    rv = data1->signal_number - data2->signal_number;

    return rv;
}

/**
 * This function is used as the comparison function by trees that hold
 * file descriptors to keep the file descriptors sorted by number. <p>
 *
 * This function returns less than zero, zero, or greater than zero to
 * indicate whether <code>key1</code> is less than, equal to, or
 * greater than <code>key2</code> respectively.
 *
 * @param[in] key1 first key
 * @param[in] key2 second key
 * @return comparison of <code>key1</code> with <code>key2</code>
 */
static int
event_loop__compare_fds(const void* key1,
                        const void* key2)
{
    /* Assume equality. */
    int rv = 0;
    event_loop__fd_data_t data1 = NULL;
    event_loop__fd_data_t data2 = NULL;

    assert(key1);
    assert(key2);

    data1 = (event_loop__fd_data_t)key1;
    data2 = (event_loop__fd_data_t)key2;

    /* Compare the file descriptors. */
    rv = data1->fd - data2->fd;

    return rv;
}

/**
 * This method guarantees that a node for <code>tree</code> exists and
 * is being held in reserve until it is needed. <p>
 *
 * The purpose of this method is to allow the methods that add
 * signals, timeouts, or file descriptors to be able to synchronously
 * return errors to their callers if those methods have to queue the
 * request because <code>tree</code> is currently busy. <p>
 *
 * By allocating and reserving the node early, these methods can set
 * things up so that it knows for sure that the request will be
 * successfully completed later. <p>
 *
 * Note, there is no corresponding method to decrement the node
 * reserve.  Instead, we just set the reserve to <code>0</code> when
 * we want to make all the reserved nodes available for use.
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 */
static void
event_loop__increment_node_reserve_for_tree(cgul_exception_t* cex,
                                            cgul_rbtree_t tree)
{
    unsigned long int reserve = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(tree);

    /* Get the current reserve count. */
    reserve = cgul_rbtree__get_cache_reserve(cex, tree);

    /* Increment the reserve count. */
    ++reserve;

    /* Make sure one more node is allocated and reserved. */
    cgul_rbtree__set_cache_reserve(cex, tree, reserve);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * This method uses <code>f</code> and <code>data</code> to create a
 * new <code>event_loop__pending_data_t</code> object which is added to
 * the <code>loop->pending_list</code>. <p>
 *
 * We add nodes to the <code>loop->pending_list</code> whenever one of
 * the other lists are busy.  When that other list is no longer busy,
 * <code>event_loop__exhaust_pending_list()</code> is called which
 * causes <code>f</code> to be called with <code>data</code>.  This
 * allows the pending requests to be honored in a timely fashion.
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] f callback function
 * @param[in] data data for <code>f</code>
 */
static void
event_loop__add_to_pending_list(cgul_exception_t* cex,
                                event_loop_t loop,
                                event_loop__pending_cb_t f,
                                void* data)
{
    event_loop__pending_data_t pending_data = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(loop);
    assert(f);

    /* Create a new data structure to hold "f" and "data".
     * pending_data will be freed by
     * event_loop__exhaust_pending_list(). */
    pending_data =
        (event_loop__pending_data_t)
            cgul_libc__calloc(cex, 1, sizeof(*pending_data));
    if (*cex) {
        goto out;
    }
    pending_data->f = f;
    pending_data->data = data;

    /* Add pending_data to pending_list. */
    cgul_list__push_back(cex, loop->pending_list, pending_data);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up pending_data if an exception occurred while adding it
     * to the pending list. */
    if (*cex && pending_data) {
        free(pending_data);
    }

    return;
}

/**
 * This method replays the requests on the
 * <code>loop->pending_list</code> in the order in which they arrived.
 * As it finishes processing each pending request, it frees the memory
 * allocated for that request. <p>
 *
 * It is possible that "add" requests could cause an exception to be
 * thrown, but because exhausting the pending list is done
 * asynchronously to when the actual "add" command was received, it
 * doesn't make much sense to let those exceptions escape.  Instead,
 * the exceptions are printed to <code>stderr</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 */
static void
event_loop__exhaust_pending_list(cgul_exception_t* cex,
                                 event_loop_t loop)
{
    event_loop__pending_data_t pending_data = NULL;
    cgul_exception_t local = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(loop);

    for (;;) {

        /* Pop the data off the front of the list. */
        pending_data = (event_loop__pending_data_t)
                           cgul_list__pop_front(cex, loop->pending_list);
        if (pending_data == NULL) {
            break;
        }

        /* Call the callback that knows how to handle this request. */
        assert(pending_data->f);
        (*pending_data->f)(&local, loop, pending_data->data);
        if (local) {
            const char* msg = cgul_exception__get_message(local);
            fprintf(stderr,
                    "Error: event_loop__exhaust_pending_list: %s\n",
                    msg);
            cgul_exception__delete(local);
            local = NULL;
        }

        /* Free pending_data. */
        free(pending_data);

    }

 out:
    return;
}

/**
 * Return the maximum file descriptor in the tree.  Because the trees
 * are kept in order, the maximum file descriptor on each tree is
 * always held in the last node in the tree.  If the tree is empty, -1
 * is returned.
 *
 * @param[in] cex c-style exception
 * @param[in] tree tree of file descriptors
 * @return maximum file descriptor in the tree
 */
static int
event_loop__find_max_fd_in_tree(cgul_exception_t* cex, cgul_rbtree_t tree)
{
    int rv = -1;
    cgul_rbtree_node_t node = NULL;
    event_loop__fd_data_t data = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(tree);

    /* Get the last node in the tree. */
    node = cgul_rbtree__get_back(cex, tree);
    if (node) {
        /* The data is held as the key. */
        data = (event_loop__fd_data_t)cgul_rbtree_node__get_key(cex, node);
        assert(data);
        rv = data->fd;
    }

 out:
    return rv;
}

/**
 * Find the maximum file descriptor in all trees.
 *
 * @param[in] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @return the maximum file descriptor in all trees
 */
static int
event_loop__find_max_fd(cgul_exception_t* cex, event_loop_t loop)
{
    int rv = -1;
    int max_wset_fd = -1;
    int max_eset_fd = -1;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(loop);

    /* rset */
    rv = event_loop__find_max_fd_in_tree(cex, loop->rset_tree);

    /* wset */
    max_wset_fd = event_loop__find_max_fd_in_tree(cex, loop->wset_tree);
    if (rv < max_wset_fd) {
        rv = max_wset_fd;
    }

    /* eset */
    max_eset_fd = event_loop__find_max_fd_in_tree(cex, loop->eset_tree);
    if (rv < max_eset_fd) {
        rv = max_eset_fd;
    }

 out:
    return rv;
}

/**
 * Call the file descriptor C-style callbacks or C++-style event
 * listeners in <code>tree</code> that are set in <code>fdset</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] tree tree of file descriptors
 * @param[in] fdset set of file descriptors returned by <code>select()</code>
 * @param[out] loop <code>event_loop</code> instance
 */
static void
event_loop__call_fd_callbacks(cgul_exception_t* cex,
                              event_loop_t loop,
                              cgul_rbtree_t tree,
                              fd_set* fdset,
                              int* tree_is_busy)
{
    cgul_rbtree_node_t curr = NULL;
    event_loop__fd_data_t data = NULL;
    cgul_exception_t local = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(tree);
    assert(fdset);

    /* We cannot add or remove entries from the tree while we are
     * iterating over it.  Instead, we flag that the tree is busy so
     * that new additions and removals end up one the "pending_list".
     * When we finish iterating over the tree, we exhaust the pending
     * list by replaying the requests.
     *
     * This simple boolean is obviously not designed to make the class
     * thread-safe; rather, it is simply designed to make it so that
     * the class can be re-entered by the same thread (which happens
     * when one callback adds or removes another callback).
     */
    *tree_is_busy = 1;

    /**
     * @todo Is there a more efficient way to test whether a file
     * descriptor is ready?  We shouldn't need to iterate over the
     * entire tree checking each file descriptor.
     */

    /* Iterate over the tree calling the callbacks for the file
     * descriptors that are ready. */
    curr = cgul_rbtree__get_front(cex, tree);
    for ( ; curr != NULL ; curr = cgul_rbtree_node__get_next(cex, curr)) {

        /* The data is held as the key. */
        data = (event_loop__fd_data_t)cgul_rbtree_node__get_key(cex, curr);
        assert(data);

        /* If data->fd is ready, call its callback.  Catch any
         * exception that may escape the callback and print it to
         * stderr. */
        if (FD_ISSET(data->fd, fdset)) {

            /*
             * The file descriptor is ready.
             */

            /* Call the C-style function callback (if any). */
            if (data->cb) {
                (*data->cb)(cex, data->fd, data->cb_data);
                if (*cex) {
                    goto out;
                }
            }

#ifdef __cplusplus

            /* Call the C++-style event listener (if any). */
            if (data->obj && data->listener) {
                ((data->obj)->*(data->listener))(data->fd);
            }

#endif  /*  __cplusplus  */

        }

    }

 out:

    /* If event_loop__add_*_fd() is called while this method is
     * iterating over the tree, it puts the request to add the fd onto
     * the pending queue, and before it does so, it reserves a rbtree
     * node so that it can report any out-of-memory exceptions
     * synchronously.  We now must use those reserved nodes. */
    cgul_rbtree__set_cache_reserve(&local, tree, 0);

    /* Deal with any pending events that happened while we were busy
     * iterating over the tree. */
    event_loop__exhaust_pending_list(&local, loop);

    /* Flag that the tree is no longer busy. */
    *tree_is_busy = 0;

    return;
}

/**
 * Add <code>data</code> to <code>data->destination_tree</code> and
 * <code>data->destination_set</code>.  This is the master function
 * that handles all the similarities between adding files for which
 * the user wants to wait for readable, writable, or exceptional
 * events.  If an error occurs, an exception is thrown. <p>
 *
 * This method also updates <code>loop->max_fd</code> if it detects
 * that <code>data->fd</code> is now the maximum file descriptor. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] data data to add to tree
 */
static void
event_loop__add_fd_to_tree(cgul_exception_t* cex,
                           event_loop_t loop,
                           event_loop__fd_data_t data)
{
    int insert_rv = 0;
    fd_set* fdset = NULL;
    cgul_rbtree_t tree = NULL;
    cgul_rbtree_node_t node = NULL;
    event_loop__fd_data_t key = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(loop);
    assert(data);

    /* Get the destination tree and destination fd_set. */
    tree = data->destination_tree;
    fdset = data->destination_set;

    /* Try to insert data. */
    insert_rv = cgul_rbtree__insert(cex, tree, data, &node);
    if (*cex) {
        goto out;
    }

    /* If a callback already exists for this file descriptor, change
     * the cb and cb_data for that handler.  Also, because we failed
     * to insert "data" as a key (because we already have one for this
     * signal), we have to remember to free "data" after copying its
     * values. */
    if (insert_rv == 0) {
        key = (event_loop__fd_data_t)cgul_rbtree_node__get_key(cex, node);
        assert(key);
        key->cb = data->cb;
        key->cb_data = data->cb_data;
#ifdef __cplusplus
        key->obj = data->obj;
        key->listener = data->listener;
#endif  /*  __cplusplus  */
        free(data);
        goto out;
    }

    /* It is more efficient to calculate the max_fd on each insert and
     * removal of a new file descriptor than it is to calculate max_fd
     * before each call to select() in event_loop__start(). */
    if (loop->max_fd < data->fd) {
        loop->max_fd = data->fd;
    }

    /* Set fd in select()'s fd_set. */
    FD_SET(data->fd, fdset);

 out:
    return;
}

/**
 * This method holds the common code for adding file descriptors to
 * the read, write, and exception trees for C-style callback
 * functions.  The main purpose of this method is to determine whether
 * <code>tree</code> is busy.  If the tree is not busy, the request to
 * add <code>fd</code> is executed immediately.  If tree is busy, the
 * request to add <code>fd</code> is placed on the pending list where
 * it is scheduled to be executed as soon as it safe to do so.  Either
 * way, <coded>event_loop__add_fd_to_tree()</code> will eventually be
 * called which really adds <code>fd</code> to <code>tree</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] fd file descriptor to monitor
 * @param[in] cb callback
 * @param[in] cb_data callback data passed to <code>cb</code>
 * @param[in] tree tree to which <code>fd</code> is added
 * @param[in] fdset set to which <code>fd</code> is added
 * @param[in] tree_is_busy boolean that protects <code>tree</code>
 */
static void
event_loop__add_fd_generically_for_c(
    cgul_exception_t* cex,
    event_loop_t loop,
    int fd,
    event_loop__fd_ready_cb_t cb,
    void* cb_data,
    cgul_rbtree_t tree,
    fd_set* fdset,
    int* tree_is_busy)
{
    event_loop__fd_data_t data = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* fd cannot be negative. */
    if (fd < 0) {
        *cex = cgul_exception__new("event_loop__add_fd_generically_for_c: "
                                    "Illegal fd: %d",
                                    fd);
        goto out;
    }

    /* Allocate data. */
    data = (event_loop__fd_data_t)cgul_libc__calloc(cex, 1, sizeof(*data));
    if (*cex) {
        goto out;
    }

    /* Initialize data. */
    data->fd = fd;
    data->cb = cb;
    data->cb_data = cb_data;
    data->destination_tree = tree;
    data->destination_set = fdset;

    /* If tree is busy, add the request to the pending list. */
    if (*tree_is_busy) {

        /* Allocate and reserve a rbtree node now so we can
         * synchronously report any out-of-memory errors. */
        event_loop__increment_node_reserve_for_tree(cex, tree);

        /* Add data to the pending list. */
        event_loop__add_to_pending_list(
            cex,
            loop,
            (event_loop__pending_cb_t)&event_loop__add_fd_to_tree,
            data);

    } else {

        /* The tree is not busy so we can directly add the fd. */
        event_loop__add_fd_to_tree(cex, loop, data);

    }

 out:

    /* Clean up data if an exception occurs after we've allocated it. */
    if (*cex && data) {
        free(data);
    }

    return;
}

#ifdef __cplusplus

/**
 * This method holds the common code for adding file descriptors to
 * the read, write, and exception trees for C++-style event listeners.
 * The main purpose of this method is to determine whether
 * <code>tree</code> is busy.  If the tree is not busy, the request to
 * add <code>fd</code> is executed immediately.  If tree is busy, the
 * request to add <code>fd</code> is placed on the pending list where
 * it is scheduled to be executed as soon as it safe to do so.  Either
 * way, <coded>event_loop__add_fd_to_tree()</code> will eventually be
 * called which really adds <code>fd</code> to <code>tree</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] fd file descriptor to monitor
 * @param[in] obj object that implements the event listener
 * @param[in] listener event listener
 * @param[in] tree tree to which <code>fd</code> is added
 * @param[in] fdset set to which <code>fd</code> is added
 * @param[in] tree_is_busy boolean that protects <code>tree</code>
 */
static void
event_loop__add_fd_generically_for_cxx(
    cgul_exception_t* cex,
    event_loop_t loop,
    int fd,
    event_loop_listener* obj,
    void (event_loop_listener::*listener)(int),
    cgul_rbtree_t tree,
    fd_set* fdset,
    int* tree_is_busy)
{
    event_loop__fd_data_t data = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* fd cannot be negative. */
    if (fd < 0) {
        *cex = cgul_exception__new("event_loop__add_fd_generically_for_cxx: "
                                    "Illegal fd: %d",
                                    fd);
        goto out;
    }

    /* Allocate data. */
    data = (event_loop__fd_data_t)cgul_libc__calloc(cex, 1, sizeof(*data));
    if (*cex) {
        goto out;
    }

    /* Initialize data. */
    data->fd = fd;
    data->obj = obj;
    data->listener = listener;
    data->destination_tree = tree;
    data->destination_set = fdset;

    /* If tree is busy, add the request to the pending list. */
    if (*tree_is_busy) {

        /* Allocate and reserve a rbtree node now so we can
         * synchronously report any out-of-memory errors. */
        event_loop__increment_node_reserve_for_tree(cex, tree);

        /* Add data to the pending list. */
        event_loop__add_to_pending_list(
            cex,
            loop,
            (event_loop__pending_cb_t)&event_loop__add_fd_to_tree,
            data);

    } else {

        /* The tree is not busy so we can directly add the fd. */
        event_loop__add_fd_to_tree(cex, loop, data);

    }

 out:

    /* Clean up data if an exception occurs after we've allocated it. */
    if (*cex && data) {
        free(data);
    }

    return;
}

#endif  /*  __cplusplus  */

/**
 * Remove <code>data->fd</code> from
 * <code>data->destination_tree</code> and
 * <code>data->destination_set</code>.  This is the master function
 * that handles all the similarities between removing files for which
 * the user wanted to wait for readable, writable, or exceptional
 * events. <p>
 *
 * This method also updates <code>loop->max_fd</code> if it detects
 * that <code>data->fd</code> is no longer the maximum file
 * descriptor. <p>
 *
 * This method calls <code>free()</code> on <code>data</code> just
 * before returning to the caller.  This is necessary because calls to
 * <code>event_loop__remove_*_fd()</code> might be routed to the
 * pending list if <code>data->destination_tree</code> is busy.  Thus,
 * it is sometimes impossible for the caller to know exactly when this
 * method is called making it impossible for the caller to know when
 * it is safe to free <code>data</code>.  Thus, this method has the
 * burden of freeing <code>data</code>. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] data data to remove from tree
*/
static void
event_loop__remove_fd_from_tree(cgul_exception_t* cex,
                                event_loop_t loop,
                                event_loop__fd_data_t data)
{
    int remove_rv = 0;
    fd_set* fdset = NULL;
    cgul_rbtree_t tree = NULL;
    void* key = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(loop);
    assert(data);

    /* Get the destination tree and destination fd_set. */
    tree = data->destination_tree;
    fdset = data->destination_set;

    /* Remove the file descriptor data from the tree. */
    remove_rv = cgul_rbtree__remove(cex, tree, data, &key, NULL);
    if (!remove_rv) {
        goto out;
    }

    /* It is more efficient to calculate the max_fd on each insert and
     * removal of a new file descriptor than it is to calculate max_fd
     * before each call to select() in event_loop__start().  You must
     * do this only after you have removed the node for this file
     * descriptor. */
    if (data->fd == loop->max_fd) {
        loop->max_fd = event_loop__find_max_fd(cex, loop);
    }

    /* Remove the file descriptor from the set. */
    FD_CLR(data->fd, fdset);

 out:

    /* We have to free "data" (which was used to look up the key).
     * See Doxygen comments above for why the caller cannot free
     * "data". */
    if (data) {
        free(data);
    }

    /* We have to free the key. */
    if (key) {
        free(key);
    }

    return;
}

/**
 * This method holds the common code for removing file descriptors
 * from the read, write, and exception trees for both C-style callback
 * functions and C++ style event listeners.  The main purpose of this
 * method is to determine whether <code>tree</code> is busy.  If the
 * tree is not busy, the request to remove <code>fd</code> is executed
 * immediately.  If tree is busy, the request to remove
 * <code>fd</code> is placed on the pending list where it is scheduled
 * to be executed as soon as it safe to do so.  Either way,
 * <coded>event_loop__remove_fd_to_tree()</code> will eventually be
 * called which really removes <code>fd</code> from <code>tree</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] fd file descriptor to monitor
 * @param[in] tree tree to which <code>fd</code> is added
 * @param[in] fdset set to which <code>fd</code> is added
 * @param[in] tree_is_busy boolean that protects <code>tree</code>
 */
static void
event_loop__remove_fd_generically(
    cgul_exception_t* cex,
    event_loop_t loop,
    int fd,
    cgul_rbtree_t tree,
    fd_set* fdset,
    int* tree_is_busy)
{
    event_loop__fd_data_t data = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* fd cannot be negative. */
    if (fd < 0) {
        *cex = cgul_exception__new("event_loop__remove_fd_generically: "
                                    "Illegal fd: %d",
                                    fd);
        goto out;
    }

    /* Allocate data. */
    data = (event_loop__fd_data_t)cgul_libc__calloc(cex, 1, sizeof(*data));
    if (*cex) {
        goto out;
    }

    /* Initialize data. */
    data->fd = fd;
    data->destination_tree = tree;
    data->destination_set = fdset;

    /* If tree is busy, add the request to the pending list. */
    if (*tree_is_busy) {

        /* Add data to the pending list. */
        event_loop__add_to_pending_list(
            cex,
            loop,
            (event_loop__pending_cb_t)&event_loop__remove_fd_from_tree,
            data);

    } else {

        /* The tree is not busy so we can directly remove the fd. */
        event_loop__remove_fd_from_tree(cex, loop, data);

    }

 out:

    /* Clean up "data" if an exception is thrown after it is
     * allocated; otherwise, event_loop__remove_fd_from_tree() is
     * responsible for freeing "data" because it might be on the
     * pending list. */
    if (*cex && data) {
        free(data);
    }

    return;
}

/**
 * Delete a <code>cgul_rbtree</code>.  This function assumes that
 * both the key and value stored in the tree are shallow data
 * structures that can be deallocated with a simple call to
 * <code>free()</code>.  If the key or value is <code>NULL</code>,
 * <code>free()</code> will not be called.  This function does not
 * take a <code>cgul_exception</code> pointer as its first argument
 * because it should only be called by the destructor.
 *
 * @param[in] tree tree to be deleted
 */
static void
event_loop__delete_tree(cgul_rbtree_t tree)
{
    void* key = NULL;
    void* value = NULL;
    cgul_exception_t local = NULL;

    assert(tree);

    /* We repeatedly remove nodes from the front of the tree until the
     * tree is empty. */
    while (cgul_rbtree__remove_front(&local, tree, &key, &value)) {
        if (key) {
            free(key);
        }
        if (value) {
            free(value);
        }
    }

    /* Free the tree itself. */
    cgul_rbtree__delete(tree);
}

/**
 * Delete a list of pending events.  The caller does not pass in a
 * <code>cgul_exception</code> pointer because this function should
 * only be called by the destructor.
 *
 * @param[in] list list to be deleted
 */
static void
event_loop__delete_pending_list(cgul_list_t list)
{
    assert(list);

    /* The pending list should always be empty by the time the
     * event_loop class is destroyed.  The reason is that the nature
     * of the "pending_list" is such that it fills up only while
     * another list is being iterated over.  Then, before returning,
     * that code returns immediately exhausting the pending list.
     * Note that if the code is compiled with C++ compiler, C++ event
     * listeners could throw an exception which could prevent the
     * pending list from being exhausted. */
#if !defined(NDEBUG) && !defined(__cplusplus)
    {
        cgul_exception_t local = NULL;
        assert(cgul_list__get_size(&local, list) == 0);
    }
#endif

    /* Free the list itself. */
    cgul_list__delete(list);
}

/**
 * Convert a <code>struct timeval</code> to a <code>double</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] tv time value
 * @return <code>tv</code> converted to a <code>double</code>
 */
static double
timeval_to_double(cgul_exception_t* cex,
                  const struct timeval* tv)
{
    double rv = 0.0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = tv->tv_sec + (tv->tv_usec / 1000000.0);

 out:
    return rv;
}

/**
 * Convert a <code>double</code> to a <code>struct timeval</code>.
 * @param[in] cex c-style exception
 * @param[in] d time as a double
 * @param[out] tv <code>d</code> converted to a <code>struct timeval</code>
 */
static void
double_to_timeval(cgul_exception_t* cex,
                  const double d, struct timeval* tv)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    tv->tv_sec = (long)d;
    tv->tv_usec = (d - tv->tv_sec) * 1000000;

 out:
    return;
}

/**
 * Get the number of microseconds since the Epoch.
 *
 * @param[in,out] cex c-style exception
 * @return number of microseconds since the epoch
*/
static double
event_loop__get_time_now(cgul_exception_t* cex)
{
    double rv = 0.0;
    struct timeval tv;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (gettimeofday(&tv, NULL) == -1) {
        /* This should never happen. */
        *cex = cgul_exception__new("event_loop__get_time_now: "
                                    "gettimeofday: %s",
                                    strerror(errno));
        goto out;
    }

    rv = timeval_to_double(cex, &tv);

 out:
    return rv;
}

/**
 * Calculate the wall-clock time when the timeout expires.
 *
 * @param[in,out] cex c-style exception
 * @param[in] period time measured from now when the timeout expires
 * @return wall-clock time when timeout expires
 */
static double
event_loop__calculate_next_timeout(cgul_exception_t* cex,
                                   double period)
{
    double rv = 0.0;
    double now = 0.0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(period >= 0.0);

    now = event_loop__get_time_now(cex);
    if (*cex) {
        goto out;
    }

    rv = now + period;

 out:
    return rv;
}

/**
 * Add the timeout specified by <code>data</code> to the timeout tree. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] data data
 */
static void
event_loop__add_timeout_to_tree(cgul_exception_t* cex,
                                event_loop_t loop,
                                event_loop__timeout_data_t data)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(loop);

    /* Insert data (as a key) into timeout_tree. */
    cgul_rbtree__insert(cex, loop->timeout_tree, data, NULL);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * Remove the timeout specified by <code>cb</code> and
 * <code>cb_data</code> from the timeout list.
 *
 * @param[in] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] timeout timeout
 */
static void
event_loop__remove_timeout_from_tree(cgul_exception_t* cex,
                                     event_loop_t loop,
                                     event_loop__timeout_data_t timeout)
{
    int remove_rv = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(loop);

    /* Try to remove the timeout from the tree. */
    remove_rv = cgul_rbtree__remove(cex,
                                     loop->timeout_tree,
                                     timeout,
                                     NULL,
                                     NULL);

    /* Free the data for the timeout if it was successfully removed. */
    if (remove_rv) {
        free(timeout);
    }

 out:
    return;
}

/**
 * This function is called after the first timeout has expired, and
 * its callback has been called.  The purpose of this function is to
 * reschedule the timeout based on its period.  Because
 * <code>loop->timeout_tree</code> is kept in sorted order based on
 * when each timeout expires this may require re-inserting this
 * timeout based on its new expiration time. <p>
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 */
static void
event_loop__reschedule_first_timeout(cgul_exception_t* cex,
                                     event_loop_t loop)
{
    int remove_rv = 0;
    void* key_as_void_ptr = NULL;
    event_loop__timeout_data_t key = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(loop);

    /* Remove the timeout from the list. */
    remove_rv = cgul_rbtree__remove_front(cex,
                                           loop->timeout_tree,
                                           &key_as_void_ptr,
                                           NULL);
    if (!remove_rv) {
        goto out;
    }

    /* Reschedule the timeout. */
    key = (event_loop__timeout_data_t)key_as_void_ptr;
    key->timeout = event_loop__calculate_next_timeout(cex, key->period);
    if (*cex) {
        goto out;
    }

    /* Re-insert the timeout.  Note that because we have turned on
     * caching of nodes for loop->timeout_tree, we are guaranteed the
     * node we just removed is waiting for us.  This allows us to
     * insert without having to worry about out-of-memory
     * exceptions. */
    cgul_rbtree__insert(cex, loop->timeout_tree, key, NULL);
    if (*cex) {
        goto out;
    }

 out:
    return;
}

/**
 * Determine if the timeout that is next in line has expired.  If so,
 * invoke its callback and reschedule or delete it.  This method only
 * handles one timeout and returns.  Don't worry about other expired
 * timeouts.  The select() system call in
 * <code>event_loop__start()</code> will be set up to timeout
 * immediately if other expired timeouts are found.  This prevents
 * your timeouts from starving your file descriptor events.
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 */
static void
event_loop__handle_next_timeout_callback(cgul_exception_t* cex,
                                         event_loop_t loop)
{
    double now = 0.0;
    event_loop__timeout_data_t first_data = NULL;
    cgul_rbtree_node_t first = NULL;
    cgul_exception_t local = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(loop);

    /* We cannot add or remove entries from the tree while we are
     * iterating over it.  Instead, we flag that the tree is busy so
     * that new additions and removals end up one the "pending_list".
     * When we finish iterating over the tree, we exhaust the pending
     * list by replaying the requests.
     *
     * This simple boolean is obviously not designed to make the class
     * thread-safe; rather, it is simply designed to make it so that
     * the class can be re-entered by the same thread (which happens
     * when one callback adds or removes another callback).
     */
    loop->timeout_tree_is_busy = 1;

    /* Because the tree of timeouts is kept in order of which will
     * expire first, you only need to check the first timeout. */
    first = cgul_rbtree__get_front(cex, loop->timeout_tree);
    if (first == NULL) {
        goto out;
    }

    first_data = (event_loop__timeout_data_t)
                     cgul_rbtree_node__get_key(cex, first);

    now = event_loop__get_time_now(cex);
    if (first_data->timeout <= now) {

        /*
         * Timeout has expired.
         */

        /* Call the C-style function callback (if any). */
        if (first_data->cb) {
            (*first_data->cb)(cex, first_data, first_data->cb_data);
            if (*cex) {
                goto out;
            }
        }

#ifdef __cplusplus

        /* Call the C++-style event listener (if any). */
        if (first_data->obj && first_data->listener) {
            ((first_data->obj)->*(first_data->listener))(first_data);
        }

#endif  /*  __cplusplus  */

        /* Reschedule the timeout. */
        event_loop__reschedule_first_timeout(cex, loop);
        if (*cex) {
            goto out;
        }
    }

 out:

    /* If event_loop__add_timeout() is called while this method is
     * iterating over the timeout_tree, it puts the request to add the
     * timeout onto the pending queue, and before it does so, it
     * reserves a rbtree node so that it can report any out-of-memory
     * exceptions synchronously.  We now must use those reserved
     * nodes. */
    cgul_rbtree__set_cache_reserve(&local, loop->timeout_tree, 0);

    /* Deal with any pending events that happened while we were busy
     * iterating over the list. */
    event_loop__exhaust_pending_list(&local, loop);

    /* Flag that the list is no longer busy. */
    loop->timeout_tree_is_busy = 0;

    return;
}

/**
 * Return the period for the next timeout.  If no timeouts are
 * scheduled, return NULL.  If a timeout is needed, "tv" is filled in
 * with the right values, and "tv" is also the value returned by this
 * function.  By returning either NULL or "tv", you make it easier on
 * the caller to properly get set up for calling select().
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[out] tv timeout to use for select if a timeout is pending
 * @return <code>NULL</code> if no timeout is pending; otherwise <code>tv</code>
*/
static struct timeval*
event_loop__get_timeout_for_select(cgul_exception_t* cex,
                                   event_loop_t loop,
                                   struct timeval* tv)
{
    double now = 0.0;
    double diff = 0.0;
    struct timeval* rv = NULL;
    cgul_rbtree_node_t front = NULL;
    event_loop__timeout_data_t data = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(loop);
    assert(tv);

    /* Get the timeout at the front of the queue. */
    front = cgul_rbtree__get_front(cex, loop->timeout_tree);
    if (front == NULL) {
        goto out;
    }

    /* Fetch the timeout data for the timeout at the front of the queue. */
    data = (event_loop__timeout_data_t)cgul_rbtree_node__get_key(cex, front);

    /* Get the time now. */
    now = event_loop__get_time_now(cex);
    if (*cex) {
        goto out;
    }

    /* Calculate the timeout of select(). */
    diff = data->timeout - now;
    if (diff <= 0.0) {
        /* We currently have pending timeouts.  So, select() should
         * return immediately.  This will give select() a chance to
         * update which file descriptors are ready. */
        tv->tv_sec = 0;
        tv->tv_usec = 0;
    } else {
        double_to_timeval(cex, diff, tv);
    }

    rv = tv;

 out:
    return rv;
}


/**
 * Set the close-on-exec flag for <code>fd</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] fd file descriptor
 */
static void
event_loop__set_close_on_exec_flag(cgul_exception_t* cex,
                                   int fd)
{
    int flags;

    assert(cex);
    if (*cex) {
        goto out;
    }

    flags = fcntl(fd, F_GETFD, 0);
    if (flags == -1) {
        *cex = cgul_exception__new("event_loop__set_close_on_exec_flag: "
                                    "fcntl(F_GETFD): %s",
                                    strerror(errno));
        goto out;
    }

    flags = fcntl(fd, F_SETFD, flags | FD_CLOEXEC);
    if (flags == -1) {
        *cex = cgul_exception__new("event_loop__set_close_on_exec_flag: "
                                    "fcntl(F_SETFD): %s",
                                    strerror(errno));
        goto out;
    }

 out:
    return;
}

/**
 * Simple wrapper around <code>close()</code> to make sure it is not
 * interrupted by a Posix asynchronous signal.  Because this function
 * is usually called as part of the destructor, it requires no
 * <code>cgul_exception</code> parameter.
 *
 * @param[in] fd file descriptor
 */
static void
event_loop__close_fd(int fd)
{
    while (close(fd) < 0) {
        if (errno == EINTR) {
            continue;
        }
        break;
    }
}

/**
 * Finalize the pair of pipe file descriptors <code>pipe_fds</code>.
 * This includes close each end of the pipe and setting the file
 * descriptors to -1.  Because this function is only ever called as
 * part of the destructor, it requires no <code>cgul_exception</code>
 * parameter.
 *
 * @param[in] pipe_fds pipe file descriptors
 */
static void
event_loop__finalize_pipe(int pipe_fds[2])
{
    /* pipe_fds[0] */
    if (pipe_fds[0] >= 0) {
        event_loop__close_fd(pipe_fds[0]);
        pipe_fds[0] = -1;
    }

    /* pipe_fds[1] */
    if (pipe_fds[1] >= 0) {
        event_loop__close_fd(pipe_fds[1]);
        pipe_fds[1] = -1;
    }
}

/**
 * This method initializes the pair of pipe file descriptors
 * <code>pipe_fds</code>.  If an error occurs, the pipe file
 * descriptors are set to <code>-1</code>, and an exception is thrown.
 *
 * @param[in] cex c-style exception
 * @param[in] pipe_fds pipe file descriptors
 */
static void
event_loop__initialize_pipe(cgul_exception_t* cex,
                            int pipe_fds[2])
{
    int is_pipe_open = 0;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Assume failure. */
    pipe_fds[0] = -1;
    pipe_fds[1] = -1;

    /* Open the pipe. */
    if (pipe(pipe_fds) == -1) {
        *cex = cgul_exception__new(
                   "event_loop__initialize_pipe: pipe: %s",
                   strerror(errno));
        goto out;
    }
    is_pipe_open = 1;

    /* Set the close-on-exec flags. */
    event_loop__set_close_on_exec_flag(cex, pipe_fds[0]);
    event_loop__set_close_on_exec_flag(cex, pipe_fds[1]);
    if (*cex) {
        goto out;
    }

 out:

    /* Clean up paritially initialized pipe. */
    if (*cex && is_pipe_open) {
        event_loop__finalize_pipe(pipe_fds);
    }

    return;
}

/**
 * Posix asynchronous signal handler that converts
 * <code>signal_number</code> into a file descriptor event where it
 * can then be safely handled in a synchronous manner.
 *
 * @param[in] signal_number signal number
 */
static void
event_loop__async_to_sync_signal_handler(int signal_number)
{
    int errno_saved = 0;

    /* Remember errno because write() could alter the value of errno
     * as a side-effect which could cause problems for the interrupted
     * function if the interrupt occurs at precisely the wrong
     * moment. */
    errno_saved = errno;

    /* FIXME: It is theoretically possible to deadlock here if for
     * some reason signals are written to the pipe much faster than
     * they can be read by the main event loop.  I do not think it is
     * a practical problem however.  A portable solution could be
     * difficult. */
    while (write(event_loop__async_to_sync_pipe[1],
                 &signal_number,
                 sizeof(int)) == -1)
    {
        if (errno == EINTR) {
            continue;
        }
        /* Just let it go. */
        break;
    }

    /* Restore the original errno. */
    errno = errno_saved;
}

/**
 * Add the signal <code>data</code> to <code>loop->signal_tree</code>.
 *
 * @param[in,out] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] data signal data
 */
static void
event_loop__add_signal_to_tree(cgul_exception_t* cex,
                               event_loop_t loop,
                               event_loop__signal_data_t data)
{
    int insert_rv = 0;
    cgul_exception_t local = NULL;
    cgul_rbtree_node_t node = NULL;
    event_loop__signal_data_t key = NULL;
    struct sigaction act;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(loop);
    assert(data);

    /* Insert "data" as a key in the signal tree. */
    insert_rv = cgul_rbtree__insert(cex, loop->signal_tree, data, &node);
    if (*cex) {
        goto out;
    }

    /* If a signal handler already exists for this signal, change the
     * cb and cb_data for that handler.  Also, because we failed to
     * insert "data" as a key (because we already have one for this
     * signal), we have to remember to free "data" after copying its
     * values. */
    if (insert_rv == 0) {
        key = (event_loop__signal_data_t)cgul_rbtree_node__get_key(cex, node);
        assert(key);
        key->cb = data->cb;
        key->cb_data = data->cb_data;
#ifdef __cplusplus
        key->obj = data->obj;
        key->listener = data->listener;
#endif  /*  __cplusplus  */
        free(data);
        goto out;
    }

    /*
     * If you reach here, this is the first signal handler registered
     * with this class for this signal.  We still need to alert the
     * operating system of the change.
     */

    /* Wait until the signal handler data has been successfully added
     * to the list before changing the real signal handler for this
     * program. */
    memset(&act, 0, sizeof(act));
    act.sa_flags = 0;
    act.sa_handler = &event_loop__async_to_sync_signal_handler;
    if (sigemptyset(&act.sa_mask) == -1) {
        *cex = cgul_exception__new("event_loop__add_signal_to_tree: "
                                    "sigemptyset: %s",
                                    strerror(errno));
        goto out;
    }
    if (sigaction(data->signal_number, &act, NULL) == -1) {
        *cex = cgul_exception__new("event_loop__add_signal_to_tree: "
                                    "sigaction: %s",
                                    strerror(errno));
        goto out;
    }

 out:

    /* Un-insert data if an exception occured after insertion.  (In
     * this case, the caller is responsible for freeing "data".) */
    if (*cex && insert_rv) {
        cgul_rbtree__remove(&local, loop->signal_tree, data, NULL, NULL);
    }

    return;
}

/**
 * Remove <code>data->signal_number</code> from the signal tree.  This
 * method calls <code>free()</code> on <code>data</code> just before
 * returning to the caller.  This is necessary because calls to
 * <code>event_loop__remove_signal_handler()</code> might be routed to
 * the pending list if the signal tree is busy.  Thus, it is sometimes
 * impossible for the caller to know exactly when this method is
 * called making it impossible for the caller to know when it is safe
 * to free <code>data</code>.  Thus, this method has the burden of
 * freeing <code>data</code>. <p>
 *
 * @param[in] cex c-style exception
 * @param[in] loop <code>event_loop</code> instance
 * @param[in] data data to remove from tree
*/
static void
event_loop__remove_signal_from_tree(cgul_exception_t* cex,
                                    event_loop_t loop,
                                    event_loop__signal_data_t data)
{
    int remove_rv = 0;
    void* key = NULL;
    struct sigaction act;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(loop);
    assert(data);

    /* Remove the signal data from the tree. */
    remove_rv = cgul_rbtree__remove(cex, loop->signal_tree, data, &key, NULL);
    if (!remove_rv) {
        goto out;
    }

    /* Restore the system default handler for the signal. */
    memset(&act, 0, sizeof(act));
    act.sa_flags = 0;
    act.sa_handler = SIG_DFL;
    if (sigemptyset(&act.sa_mask) == -1) {
        *cex = cgul_exception__new("event_loop__remove_signal_from_tree: "
                                    "sigemptyset: %s",
                                    strerror(errno));
        goto out;
    }
    if (sigaction(data->signal_number, &act, NULL) == -1) {
        *cex = cgul_exception__new("event_loop__remove_signal_from_tree: "
                                    "sigaction: %s",
                                    strerror(errno));
        goto out;
    }

 out:

    /* We have to free "data" (which was used to look up the key).
     * See Doxygen comments above for why the caller cannot free
     * "data". */
    if (data) {
        free(data);
    }

    /* We have to free the key. */
    if (key) {
        free(key);
    }

    return;
}

/**
 * Read the Posix signal number out of the pipe and call the correct
 * signal handler.  The <code>fd</code> parameter is the
 * <code>event_loop__async_to_sync_pipe[0]</code> file descriptor.
 *
 * @param[in] cex c-style exception
 * @param[in] fd file descriptor 
 * @param[in] data client data pointing to the <code>event_loop</code> instance
 */
static void
event_loop__call_signal_handlers(cgul_exception_t* cex,
                                 int fd,
                                 void* data)
{
    int signal_number;
    event_loop_t loop = NULL;
    cgul_rbtree_node_t node = NULL;
    cgul_exception_t local = NULL;
    event_loop__signal_data_t key = NULL;
    struct event_loop__signal_data search_key = {0};

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(data);

    loop = (event_loop_t)data;

    /* Sanity check.  (This functions should only be registered as the
     * call back for the event_loop__async_to_sync_pipe[0] file
     * descriptor.) */
    if (fd != event_loop__async_to_sync_pipe[0]) {
        *cex = cgul_exception__new(
            "event_loop__call_signal_handlers: "
            "called for unexpected file descriptor: %d", fd);
        goto out;
    }

    /* Determine which signal was raised. */
    while (read(fd, &signal_number, sizeof(int)) == -1) {
        if (errno == EINTR) {
            continue;
        }
        *cex = cgul_exception__new("event_loop__call_signal_handlers: "
                                    "read: %s",
                                    strerror(errno));
        goto out;
    }

    /* We cannot add or remove entries from the tree while we are
     * iterating over it.  Instead, we flag that the tree is busy so
     * that new additions and removals end up one the "pending_list".
     * When we finish iterating over the tree, we exhaust the pending
     * list by replaying the requests.
     *
     * This simple boolean is obviously not designed to make the class
     * thread-safe; rather, it is simply designed to make it so that
     * the class can be re-entered by the same thread (which happens
     * when one callback adds or removes another callback).
     */
    loop->signal_tree_is_busy = 1;

    /* Initialize the key we use for searching the tree.  Only the
     * signal_number is relevant. */
    search_key.signal_number = signal_number;

    /* Find the node in the tree for this signal. */
    node = cgul_rbtree__find(cex, loop->signal_tree, &search_key);

    /* If we could not find the node, a signal has not been registered. */
    if (node == NULL) {
        goto out;
    }

    /* Get the key from the node. */
    key = (event_loop__signal_data_t)cgul_rbtree_node__get_key(cex, node);

    /* Call the C-style function callback (if any). */
    if (key->cb) {
        (*key->cb)(cex, signal_number, key->cb_data);
        if (*cex) {
            goto out;
        }
    }

#ifdef __cplusplus

    /* Call the C++-style event listener (if any). */
    if (key->obj && key->listener) {
        ((key->obj)->*(key->listener))(signal_number);
    }

#endif  /*  __cplusplus  */

 out:

    /* If event_loop__add_signal_handler() is called while this method is
     * iterating over the siganl_tree, it puts the request to add the
     * signal handler onto the pending queue, and before it does so, it
     * reserves a rbtree node so that it can report any out-of-memory
     * exceptions synchronously.  We now must use those reserved
     * nodes. */
    cgul_rbtree__set_cache_reserve(&local, loop->signal_tree, 0);

    /* Deal with any pending events that happened while we were busy
     * iterating over the list. */
    event_loop__exhaust_pending_list(&local, loop);

    /* Flag that the list is no longer busy. */
    loop->signal_tree_is_busy = 0;

    return;
}

/**
 * Read the internal message out of the internal pipe and act
 * accordingly.  The internal messages are defined near the top of
 * this file as "static" integers.  The <code>fd</code> parameter is
 * the <code>internal_pipe</code> file descriptor for the event loop
 * which is pass as the client data <code>data</code>.
 *
 * @param[in] cex c-style exception
 * @param[in] fd file descriptor 
 * @param[in] data client data pointing to the <code>event_loop</code> instance
 */
static void
event_loop__read_internal_message(cgul_exception_t* cex,
                                  int fd,
                                  void* data)
{
    int message_number = 0;
    event_loop_t loop = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(data);

    loop = (event_loop_t)data;

    /* Sanity check.  (This functions should only be registered as the
     * call back for the loop->internal_pipe[0] file descriptor.) */
    if (fd != loop->internal_pipe[0]) {
        *cex = cgul_exception__new(
            "event_loop__read_internal_message: "
            "called for unexpected file descriptor: %d", fd);
        goto out;
    }

    /* Read the internal message. */
    while (read(fd, &message_number, sizeof(int)) == -1) {
        if (errno == EINTR) {
            continue;
        }
        *cex = cgul_exception__new(
            "event_loop__read_internal_message: read: %m");
        goto out;
    }

    /* Handle the internal message. */
    switch (message_number) {
    case EVENT_LOOP__INTERNAL_STOP_MESSAGE:
        /* Stop the event loop. */
        loop->stop = 1;
        break;
    default:
        *cex = cgul_exception__new(
            "event_loop__read_internal_message: "
            "invalid message number: %d", message_number);
        goto out;
    }

 out:
    return;
}

event_loop_t
event_loop__get_instance(cgul_exception_t* cex)
{
    event_loop_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    if (event_loop__instance == NULL) {

        /* Create the singleton. */
        event_loop__instance = event_loop__new(cex);
        if (*cex) {
            goto out;
        }

        /* Stock instances created with "event_loop__new()" (as above)
         * are not configured to handle Posix asynchronous signals
         * (because the only way to pass in the file descriptors for
         * the pipe to the Posix asynchronous signal handler is
         * through a "static" variable).  Thus, we still need to set
         * up the async_to_sync pipe so that
         * event_loop__add_signal_handler() works with the
         * singleton. */
        event_loop__initialize_pipe(cex, event_loop__async_to_sync_pipe);
        event_loop__add_read_fd(cex,
                                event_loop__instance,
                                event_loop__async_to_sync_pipe[0],
                                &event_loop__call_signal_handlers,
                                (void*)event_loop__instance);
        if (*cex) {
            goto out;
        }

    }

    /* Return the singelton. */
    rv = event_loop__instance;

 out:
    return rv;
}

extern void
event_loop__put_instance()
{
    /* event_loop__instance */
    if (event_loop__instance) {
        event_loop__delete(event_loop__instance);
        event_loop__instance = NULL;
    }

    /* Close async-to-sync pipe. */
    event_loop__finalize_pipe(event_loop__async_to_sync_pipe);
}

event_loop_t
event_loop__new(cgul_exception_t* cex)
{
    event_loop_t rv = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    rv = (event_loop_t)cgul_libc__calloc(cex, 1, sizeof(*rv));
    if (*cex) {
        goto out;
    }

    /*
     * Construct the object.
     */

    /* Pipe used for passing thread-safe internal messages. */
    event_loop__initialize_pipe(cex, rv->internal_pipe);

    /* Create the timeout_tree and enabling caching of nodes for the
     * timeout tree.  This makes it possible to reuse nodes without
     * having to worry about malloc() failing when we reschedule
     * timeouts.  Without caching the nodes, this would thrash
     * malloc() and require that we have some sort of asynchronous
     * error reporting mechanism if malloc() returns an out-of-memory
     * exception. */
    rv->timeout_tree = cgul_rbtree__new(cex, &event_loop__compare_timeouts);
    cgul_rbtree__set_cache_size(cex, rv->timeout_tree, 16, /* preallocate */ 1);

    /* Create the signal_tree and enabling caching of nodes for the
     * same reason we enable caching nodes for the timeout tree.*/
    rv->signal_tree = cgul_rbtree__new(cex, &event_loop__compare_signals);
    cgul_rbtree__set_cache_size(cex, rv->signal_tree,  16, /* preallocate */ 1);

    /* Create the rset_tree and enabling caching of nodes for the same
     * reason we enable caching nodes for the timeout tree.*/
    rv->rset_tree = cgul_rbtree__new(cex, &event_loop__compare_fds);
    cgul_rbtree__set_cache_size(cex, rv->rset_tree,  16, /* preallocate */ 1);

    /* Create the wset_tree and enabling caching of nodes for the same
     * reason we enable caching nodes for the timeout tree.*/
    rv->wset_tree = cgul_rbtree__new(cex, &event_loop__compare_fds);
    cgul_rbtree__set_cache_size(cex, rv->wset_tree,  16, /* preallocate */ 1);

    /* Create the eset_tree and enabling caching of nodes for the same
     * reason we enable caching nodes for the timeout tree.*/
    rv->eset_tree = cgul_rbtree__new(cex, &event_loop__compare_fds);
    cgul_rbtree__set_cache_size(cex, rv->eset_tree,  16, /* preallocate */ 1);

    /* Create the queue of pending commands that are waiting for a
     * tree to no longer be busy before they can be executed. */
    rv->pending_list = cgul_list__new(cex);
    cgul_list__set_cache_size(cex, rv->pending_list,  16, /* preallocate */ 1);

    if (*cex) {
        goto out;
    }

    FD_ZERO(&rv->rset);
    FD_ZERO(&rv->wset);
    FD_ZERO(&rv->eset);

    /* At this point the object is fully constructed.  So, you can
     * safely start calling methods which is good because we need to
     * register ourself to read messages out of the internal pipe so
     * that "event_loop__stop()" (for example) can do its job. */
    event_loop__add_read_fd(cex,
                            rv,
                            rv->internal_pipe[0],
                            &event_loop__read_internal_message,
                            (void*)rv);
    if (*cex) {
        goto out;
    }

 out:

    /* Delete partially constructed objects. */
    if (*cex && rv) {
        event_loop__delete(rv);
        rv = NULL;
    }

    return rv;
}

void
event_loop__delete(event_loop_t loop)
{
    if (loop) {

        /* Delete the pending list. */
        if (loop->pending_list) {
            event_loop__delete_pending_list(loop->pending_list);
        }

        /* Delete the trees. */
        if (loop->rset_tree) {
            event_loop__delete_tree(loop->rset_tree);
        }
        if (loop->wset_tree) {
            event_loop__delete_tree(loop->wset_tree);
        }
        if (loop->eset_tree) {
            event_loop__delete_tree(loop->eset_tree);
        }
        if (loop->signal_tree) {
            event_loop__delete_tree(loop->signal_tree);
        }
        if (loop->timeout_tree) {
            event_loop__delete_tree(loop->timeout_tree);
        }

        /* internal_pipe */
        event_loop__finalize_pipe(loop->internal_pipe);

        /* Free the object itself. */
        free(loop);
    }
}

void
event_loop__start(cgul_exception_t* cex,
                  event_loop_t loop)
{
    int n = -1;
    int select_rv = -1;
    struct timeval* t = NULL;
    struct timeval tv;
    fd_set rset;
    fd_set wset;
    fd_set eset;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(loop);

    for (;;) {

        /* Check to see if the client wants to stop processing events. */
        if (loop->stop) {
            goto out;
        }

        /* Determine the timeout needed for select(). */
        t = event_loop__get_timeout_for_select(cex, loop, &tv);
        if (*cex) {
            goto out;
        }

        /* Prepare the file descriptor sets. */
        n = loop->max_fd + 1;
        memcpy(&rset, &loop->rset, sizeof(rset));
        memcpy(&wset, &loop->wset, sizeof(wset));
        memcpy(&eset, &loop->eset, sizeof(eset));

        /* Wait. */
        select_rv = select(n, &rset, &wset, &eset, t);
        if (select_rv == -1) {
            if (errno == EINTR) {
                /* You need to reinitialize the fd_set values and
                 * check for timeouts. */
                continue;
            }
            *cex = cgul_exception__new("event_loop__start: select: %s\n",
                                        strerror(errno));
            goto out;
        }

        /*
         * At least one event has arrived.
         */

        /* Check for file descriptor events.  Note that because
         * signals are written to the async-to-sync pipe and because
         * the constructor for this class registers a callback for the
         * pipe, signals are handled when the event loop detects that
         * the pipe is readable. */
        if (select_rv > 0) {
            event_loop__call_fd_callbacks(cex,
                                          loop,
                                          loop->rset_tree,
                                          &rset,
                                          &loop->rset_tree_is_busy);
            event_loop__call_fd_callbacks(cex,
                                          loop,
                                          loop->wset_tree,
                                          &wset,
                                          &loop->wset_tree_is_busy);
            event_loop__call_fd_callbacks(cex,
                                          loop,
                                          loop->eset_tree,
                                          &eset,
                                          &loop->eset_tree_is_busy);
        }

        /* Handle the next timeout on the queue if it has expired. */
        event_loop__handle_next_timeout_callback(cex, loop);

        /* Propagate exceptions. */
        if (*cex) {
            goto out;
        }
    }

 out:
    return;
}

void
event_loop__stop(cgul_exception_t* cex,
                 event_loop_t loop)
{
    int message = EVENT_LOOP__INTERNAL_STOP_MESSAGE;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(loop);

    /* If the main loop is blocked on select(), wake it up by writing
     * the EVENT_LOOP__INTERNAL_STOP_MESSAGE message to the internal
     * pipe.  This event is handled by the
     * "event_loop__read_internal_message()" function which will set
     * loop->stop to true. */
    while (write(loop->internal_pipe[1], &message, sizeof(message)) == -1) {
        if (errno == EINTR) {
            continue;
        }
        *cex = cgul_exception__new("event_loop__stop: write: %m");
        goto out;
    }

 out:
    return;
}

void
event_loop__add_signal_handler(cgul_exception_t* cex,
                               event_loop_t loop,
                               int signal_number,
                               event_loop__signal_cb_t cb,
                               void* cb_data)
{
    event_loop__signal_data_t data = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(loop);

    /* Allocate space for the data. */
    data = (event_loop__signal_data_t)cgul_libc__calloc(cex, 1, sizeof(*data));
    if (*cex) {
        goto out;
    }

    /* Initialize the data. */
    data->signal_number = signal_number;
    data->cb = cb;
    data->cb_data = cb_data;

    /* If signal_tree is busy, add the request to the pending list. */
    if (loop->signal_tree_is_busy) {

        /* Allocate and reserve a rbtree node now so we can
         * synchronously report any out-of-memory errors. */
        event_loop__increment_node_reserve_for_tree(cex, loop->signal_tree);

        /* Add data to the pending list. */
        event_loop__add_to_pending_list(
            cex,
            loop,
            (event_loop__pending_cb_t)&event_loop__add_signal_to_tree,
            data);

    } else {

        /* The signal_tree is not busy so we can directly add the signal. */
        event_loop__add_signal_to_tree(cex, loop, data);

    }

 out:

    /* Free data if an error occurs after it is allocated. */
    if (*cex && data) {
        free(data);
    }

    return;
}

#ifdef __cplusplus

void
event_loop__add_signal_handler_listener(
    cgul_exception_t* cex,
    event_loop_t loop,
    int signal_number,
    event_loop_listener* obj,
    void (event_loop_listener::*listener)(int))
{
    event_loop__signal_data_t data = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(loop);

    /* Allocate space for the data. */
    data = (event_loop__signal_data_t)cgul_libc__calloc(cex, 1, sizeof(*data));
    if (*cex) {
        goto out;
    }

    /* Initialize the data. */
    data->signal_number = signal_number;
    data->obj = obj;
    data->listener = listener;

    /* If signal_tree is busy, add the request to the pending list. */
    if (loop->signal_tree_is_busy) {

        /* Allocate and reserve a rbtree node now so we can
         * synchronously report any out-of-memory errors. */
        event_loop__increment_node_reserve_for_tree(cex, loop->signal_tree);

        /* Add data to the pending list. */
        event_loop__add_to_pending_list(
            cex,
            loop,
            (event_loop__pending_cb_t)&event_loop__add_signal_to_tree,
            data);

    } else {

        /* The signal_tree is not busy so we can directly add the signal. */
        event_loop__add_signal_to_tree(cex, loop, data);

    }

 out:

    /* Free data if an error occurs after it is allocated. */
    if (*cex && data) {
        free(data);
    }

    return;
}

#endif  /*  __cplusplus  */

void
event_loop__remove_signal_handler(cgul_exception_t* cex,
                                  event_loop_t loop,
                                  int signal_number)
{
    event_loop__signal_data_t data = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    /* Allocate data. */
    data = (event_loop__signal_data_t)cgul_libc__calloc(cex, 1, sizeof(*data));
    if (*cex) {
        goto out;
    }

    /* Initialize data. */
    data->signal_number = signal_number;

    /* If tree is busy, add the request to the pending list. */
    if (loop->signal_tree_is_busy) {

        /* Add data to the pending list. */
        event_loop__add_to_pending_list(
            cex,
            loop,
            (event_loop__pending_cb_t)&event_loop__remove_signal_from_tree,
            data);

    } else {

        /* The tree is not busy so we can directly remove the fd. */
        event_loop__remove_signal_from_tree(cex, loop, data);

    }

 out:

    /* Clean up "data" if an exception is thrown after it is
     * allocated; otherwise, event_loop__remove_signal_from_tree() is
     * responsible for freeing "data" because it might be on the
     * pending list. */
    if (*cex && data) {
        free(data);
    }

    return;
}

event_loop__timeout_t
event_loop__add_timeout(cgul_exception_t* cex,
                        event_loop_t loop,
                        double period,
                        event_loop__timeout_cb_t cb,
                        void* cb_data)
{
    event_loop__timeout_t rv = NULL;
    event_loop__timeout_data_t data = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(loop);

    /* Timeouts with a negative period are not added to the tree. */
    if (period < 0.0) {
        *cex = cgul_exception__new("event_loop__add_timeout: "
                                   "timeout period must be non-negative");
        goto out;
    }

    /* Timeouts with a zero period are set to a small value (relative
     * to the typical computer clock resolution) to prevent the zero
     * second timeout from starving the others. */
    if (period == 0.0) {
        period = 1.0e-12;
    }

    /* Allocate the data now so we can return a synchronous error even
     * if we have to queue the request because the tree is busy. */
    data = (event_loop__timeout_data_t)
               cgul_libc__calloc(cex, 1, sizeof(*data));
    if (*cex) {
        goto out;
    }

    /* Initialize data. */
    data->period = period;
    data->timeout = event_loop__calculate_next_timeout(cex, period);
    data->cb = cb;
    data->cb_data = cb_data;

    /* If the timeout_tree is busy, add the request to the pending list. */
    if (loop->timeout_tree_is_busy) {

        /* Allocate and reserve a rbtree node now so we can
         * synchronously report any out-of-memory errors. */
        event_loop__increment_node_reserve_for_tree(cex, loop->timeout_tree);

        /* Add data to the pending list. */
        event_loop__add_to_pending_list(
            cex,
            loop,
            (event_loop__pending_cb_t)&event_loop__add_timeout_to_tree,
            data);

    } else {

        /* The timeout_tree is not busy so we can directly add it. */
        event_loop__add_timeout_to_tree(cex, loop, data);

    }

    if (*cex) {
        goto out;
    }

    /*
     * Success.
     */

    /* We have inserted or will soon insert "data" as a key in
     * loop->timeout_tree.  We return the key as an opaque data
     * pointer to the client so that the client can pass it back to us
     * when the client wants to remove the timeout. */
    rv = (event_loop__timeout_t)data;

 out:

    /* Clean up data if an exception occurs after we've allocated it. */
    if (*cex && data) {
        free(data);
    }

    return rv;
}

#ifdef __cplusplus

event_loop__timeout_t
event_loop__add_timeout_listener(
    cgul_exception_t* cex,
    event_loop_t loop,
    double period,
    event_loop_listener* obj,
    void (event_loop_listener::*listener)(event_loop__timeout_t))
{
    event_loop__timeout_t rv = NULL;
    event_loop__timeout_data_t data = NULL;

    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(loop);

    /* Timeouts with a negative period are not added to the tree. */
    if (period < 0.0) {
        *cex = cgul_exception__new("event_loop__add_timeout: "
                                   "timeout period must be non-negative");
        goto out;
    }

    /* Timeouts with a zero period are set to a small value (relative
     * to the typical computer clock resolution) to prevent the zero
     * second timeout from starving the others. */
    if (period == 0.0) {
        period = 1.0e-12;
    }

    /* Allocate the data now so we can return a synchronous error even
     * if we have to queue the request because the tree is busy. */
    data = (event_loop__timeout_data_t)
               cgul_libc__calloc(cex, 1, sizeof(*data));
    if (*cex) {
        goto out;
    }

    /* Initialize data. */
    data->period = period;
    data->timeout = event_loop__calculate_next_timeout(cex, period);
    data->obj = obj;
    data->listener = listener;

    /* If the timeout_tree is busy, add the request to the pending list. */
    if (loop->timeout_tree_is_busy) {

        /* Allocate and reserve a rbtree node now so we can
         * synchronously report any out-of-memory errors. */
        event_loop__increment_node_reserve_for_tree(cex, loop->timeout_tree);

        /* Add data to the pending list. */
        event_loop__add_to_pending_list(
            cex,
            loop,
            (event_loop__pending_cb_t)&event_loop__add_timeout_to_tree,
            data);

    } else {

        /* The timeout_tree is not busy so we can directly add it. */
        event_loop__add_timeout_to_tree(cex, loop, data);

    }

    if (*cex) {
        goto out;
    }

    /*
     * Success.
     */

    /* We have inserted or will soon insert "data" as a key in
     * loop->timeout_tree.  We return the key as an opaque data
     * pointer to the client so that the client can pass it back to us
     * when the client wants to remove the timeout. */
    rv = (event_loop__timeout_t)data;

 out:

    /* Clean up data if an exception occurs after we've allocated it. */
    if (*cex && data) {
        free(data);
    }

    return rv;
}

#endif  /*  __cplusplus  */

void
event_loop__remove_timeout(cgul_exception_t* cex,
                           event_loop_t loop,
                           event_loop__timeout_t timeout)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    assert(loop);

    /* If the timeout_tree is busy, add the request to the pending list. */
    if (loop->timeout_tree_is_busy) {

        event_loop__add_to_pending_list(
            cex,
            loop,
            (event_loop__pending_cb_t)&event_loop__remove_timeout_from_tree,
            timeout);

    } else {

        /* The timeout_tree is not busy so we can directly remove it. */
        event_loop__remove_timeout_from_tree(
            cex,
            loop,
            (event_loop__timeout_data_t)timeout);

    }

 out:
    return;
}

void
event_loop__add_read_fd(cgul_exception_t* cex,
                        event_loop_t loop,
                        int fd,
                        event_loop__fd_ready_cb_t cb,
                        void* cb_data)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    event_loop__add_fd_generically_for_c(
        cex,
        loop,
        fd,
        cb,
        cb_data,
        loop->rset_tree,
        &loop->rset,
        &loop->rset_tree_is_busy);

 out:
    return;
}

#ifdef __cplusplus

void
event_loop__add_read_fd_listener(
    cgul_exception_t* cex,
    event_loop_t loop,
    int fd,
    event_loop_listener* obj,
    void (event_loop_listener::*listener)(int))
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    event_loop__add_fd_generically_for_cxx(
        cex,
        loop,
        fd,
        obj,
        listener,
        loop->rset_tree,
        &loop->rset,
        &loop->rset_tree_is_busy);

 out:
    return;
}

#endif  /*  __cplusplus  */

void
event_loop__remove_read_fd(cgul_exception_t* cex,
                           event_loop_t loop,
                           int fd)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    event_loop__remove_fd_generically(
        cex,
        loop,
        fd,
        loop->rset_tree,
        &loop->rset,
        &loop->rset_tree_is_busy);

 out:
    return;
}

void
event_loop__add_write_fd(cgul_exception_t* cex,
                         event_loop_t loop,
                         int fd,
                         event_loop__fd_ready_cb_t cb,
                         void* cb_data)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    event_loop__add_fd_generically_for_c(
        cex,
        loop,
        fd,
        cb,
        cb_data,
        loop->wset_tree,
        &loop->wset,
        &loop->wset_tree_is_busy);

 out:
    return;
}

#ifdef __cplusplus

void
event_loop__add_write_fd_listener(
    cgul_exception_t* cex,
    event_loop_t loop,
    int fd,
    event_loop_listener* obj,
    void (event_loop_listener::*listener)(int))
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    event_loop__add_fd_generically_for_cxx(
        cex,
        loop,
        fd,
        obj,
        listener,
        loop->wset_tree,
        &loop->wset,
        &loop->wset_tree_is_busy);

 out:
    return;
}

#endif  /*  __cplusplus  */

void
event_loop__remove_write_fd(cgul_exception_t* cex,
                            event_loop_t loop,
                            int fd)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    event_loop__remove_fd_generically(
        cex,
        loop,
        fd,
        loop->wset_tree,
        &loop->wset,
        &loop->wset_tree_is_busy);

 out:
    return;
}

void
event_loop__add_exception_fd(cgul_exception_t* cex,
                             event_loop_t loop,
                             int fd,
                             event_loop__fd_ready_cb_t cb,
                             void* cb_data)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    event_loop__add_fd_generically_for_c(
        cex,
        loop,
        fd,
        cb,
        cb_data,
        loop->eset_tree,
        &loop->eset,
        &loop->eset_tree_is_busy);

 out:
    return;
}

#ifdef __cplusplus

void
event_loop__add_exception_fd_listener(
    cgul_exception_t* cex,
    event_loop_t loop,
    int fd,
    event_loop_listener* obj,
    void (event_loop_listener::*listener)(int))
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    event_loop__add_fd_generically_for_cxx(
        cex,
        loop,
        fd,
        obj,
        listener,
        loop->eset_tree,
        &loop->eset,
        &loop->eset_tree_is_busy);

 out:
    return;
}

#endif  /*  __cplusplus  */

void
event_loop__remove_exception_fd(cgul_exception_t* cex,
                                event_loop_t loop,
                                int fd)
{
    assert(cex);
    if (*cex) {
        goto out;
    }

    event_loop__remove_fd_generically(
        cex,
        loop,
        fd,
        loop->eset_tree,
        &loop->eset,
        &loop->eset_tree_is_busy);

 out:
    return;
}
