#include "event_loop_cxx.h"
#include <signal.h>
#include <iostream>
#include <stdexcept>
#include <string>

using namespace std;

class Listeners : public event_loop_listener {

public:

    /* NOTE: The event loop can easily be set up to convert asynchronous
     * signal handlers into synchronous signal handlers. */
    virtual void
    handle_sigint(int signal_number)
    {
        /* Print message. */
        printf("SIGINT (%d) caught.  Exiting!\n", signal_number);
        /* Stop the event loop. */
        event_loop_cxx::get_instance()->stop();
    }

    virtual void
    stdin_readable(int fd)
    {
        string line;

        /* Get the next line of input. */
        getline(cin, line);
        if (!cin) {

            if (cin.eof()) {
                /* EOF -- Remove listener. */
                event_loop_cxx::get_instance()->remove_read_fd(fd);
            } else {
                /* Error */
                throw runtime_error("Unable to read from stdin!");
            }

        } else if (line == "exit") {

            /* Stop the event loop so we can exit. */
            event_loop_cxx::get_instance()->stop();

        } else if (line == "disconnect") {

            /* Test re-entrant code by removing this callback from the
             * callback itself. */
            event_loop_cxx::get_instance()->remove_read_fd(fd);

        } else {

            /* Echo string back to user. */
            cout << line << endl;

        }
    }

    virtual void
    rapid(event_loop_cxx::timeout_t timeout)
    {
        static int count = 0;
        ++count;
        cout << "rapid (" << count << " of 10)" << endl << flush;
        if (count >= 10) {
            event_loop_cxx::get_instance()->remove_timeout(timeout);
        }
    }

    virtual void
    tick(event_loop_cxx::timeout_t timeout)
    {
        static bool first_pass = true;
        cout << "tick: type something!" << endl << flush;
        if (first_pass) {
            /* Add the "rapid" timeout if this is the first pass. */
            first_pass = false;
            event_loop_cxx::get_instance()->add_timeout(
                0.25, this,
                EVENT_LOOP__UPCAST_TIMEOUT(&Listeners::rapid));
        }
    }

    virtual void
    tock(event_loop_cxx::timeout_t timeout)
    {
        cout << "tock: type something!" << endl << flush;
    }

};

int
main(int argc,
     char* argv[])
{
    int rv = 0;
    event_loop_cxx* loop = NULL;

    try {

        /* Get the event_loop instance. */
        loop = event_loop_cxx::get_instance();

        /* Create the object that listens for events. */
        Listeners listeners;

        /* Add events you are interested in receiving .*/
        loop->add_signal_handler(
            SIGINT, &listeners,
            EVENT_LOOP__UPCAST_FD_READY(&Listeners::handle_sigint));
        loop->add_read_fd(
            fileno(stdin), &listeners,
            EVENT_LOOP__UPCAST_FD_READY(&Listeners::stdin_readable));
        loop->add_timeout(
            2.0, &listeners,
            EVENT_LOOP__UPCAST_TIMEOUT(&Listeners::tick));
        loop->add_timeout(
            6.5, &listeners,
            EVENT_LOOP__UPCAST_TIMEOUT(&Listeners::tock));

        /* Enter the main loop. */
        loop->start();

    } catch (const exception& e) {
        cerr << "*** Error: " << e.what() << endl;
        rv = 1;
    }

    /* Clean up. */
    if (loop) {
        loop->put_instance();
    }

    return rv;
}
